@@ -2,15 +2,15 @@ import logging
 from concurrent.futures import Executor, Future, ProcessPoolExecutor
 from enum import Enum
 from threading import Lock
-from typing import Callable, Dict
+from typing import Callable, Dict, List, Union
 
 import numpy as np
-
 import pandas as pd
-from pydantic import BaseModel, Field, root_validator
+from pandas import DataFrame
+from pydantic import ConfigDict, Field, model_validator
 
 from xopt.errors import XoptError
-from xopt.pydantic import JSON_ENCODERS, NormalExecutor
+from xopt.pydantic import NormalExecutor, XoptBaseModel
 from xopt.utils import get_function, get_function_defaults, safe_call
 
 logger = logging.getLogger(__name__)
@@ -22,7 +22,7 @@ class ExecutorEnum(str, Enum):
     normal_executor = "NormalExecutor"
 
 
-class Evaluator(BaseModel):
+class Evaluator(XoptBaseModel):
     """
     Xopt Evaluator for handling the parallel execution of an evaluate function.
 
@@ -37,27 +37,34 @@ class Evaluator(BaseModel):
     executor : NormalExecutor
         NormalExecutor or any instantiated Executor object
     vectorized : bool, default=False
-        If true,
+        If true, lists of evaluation points will be sent to the evaluator
+        function to be processed in parallel instead of evaluated separately via
+        mapping.
     """
 
     function: Callable
-    max_workers: int = 1
+    max_workers: int = Field(1, ge=1)
     executor: NormalExecutor = Field(exclude=True)  # Do not serialize
-    function_kwargs: dict = {}
-    vectorized: bool = False
+    function_kwargs: dict = Field({})
+    vectorized: bool = Field(False)
 
-    class Config:
-        """config"""
+    model_config = ConfigDict(arbitrary_types_allowed=True)
 
-        arbitrary_types_allowed = True
-        # validate_assignment = True # Broken in 1.9.0.
-        # Trying to fix in https://github.com/samuelcolvin/pydantic/pull/4194
-        json_encoders = JSON_ENCODERS
-        extra = "forbid"
-        # copy_on_model_validation = False
+    @model_validator(mode="before")
+    def validate_all(cls, values: Dict) -> Dict:
+        """
+        Validate all inputs before initializing the Evaluator.
 
-    @root_validator(pre=True)
-    def validate_all(cls, values):
+        Parameters
+        ----------
+        values : dict
+            The input values to validate.
+
+        Returns
+        -------
+        dict
+            The validated input values.
+        """
         f = get_function(values["function"])
         kwargs = values.get("function_kwargs", {})
         kwargs = {**get_function_defaults(f), **kwargs}
@@ -68,10 +75,11 @@ class Evaluator(BaseModel):
 
         executor = values.pop("executor", None)
         if not executor:
-            if max_workers > 1:
-                executor = ProcessPoolExecutor(max_workers=max_workers)
-            else:
-                executor = DummyExecutor()
+            executor = (
+                ProcessPoolExecutor(max_workers=max_workers)
+                if max_workers > 1
+                else DummyExecutor()
+            )
 
         # Cast as a NormalExecutor
         values["executor"] = NormalExecutor[type(executor)](executor=executor)
@@ -79,75 +87,131 @@ class Evaluator(BaseModel):
 
         return values
 
-    def evaluate(self, input: Dict, **kwargs):
+    def evaluate(self, input: Dict, **kwargs) -> Dict:
         """
         Evaluate a single input dict using Evaluator.function with
         Evaluator.function_kwargs.
 
         Further kwargs are passed to the function.
 
-        Inputs:
-            inputs: dict of inputs to be evaluated
-            **kwargs: additional kwargs to pass to the function
-
-        Returns:
-            function(input, **function_kwargs_updated)
+        Parameters
+        ----------
+        input : dict
+            The input dictionary to evaluate.
+        **kwargs : dict
+            Additional keyword arguments to pass to the function.
 
+        Returns
+        -------
+        dict
+            The evaluation result.
         """
         return self.safe_function(input, **{**self.function_kwargs, **kwargs})
 
-    def evaluate_data(self, input_data: pd.DataFrame):
-        """evaluate dataframe of inputs"""
-        input_data = pd.DataFrame(input_data)
+    def evaluate_data(
+        self,
+        input_data: Union[
+            pd.DataFrame,
+            List[Dict[str, float]],
+            Dict[str, List[float]],
+            Dict[str, float],
+        ],
+    ) -> pd.DataFrame:
+        """
+        Evaluate a dataframe of inputs.
 
+        Parameters
+        ----------
+        input_data : Union[pd.DataFrame, List[Dict[str, float]], Dict[str, List[float]], Dict[str, float]]
+            The input data to evaluate.
+
+        Returns
+        -------
+        pd.DataFrame
+            The evaluation results.
+        """
         if self.vectorized:
             output_data = self.safe_function(input_data, **self.function_kwargs)
         else:
             # This construction is needed to avoid a pickle error
+            # translate input data into pandas dataframes
+            if not isinstance(input_data, DataFrame):
+                try:
+                    input_data = DataFrame(input_data)
+                except ValueError:
+                    input_data = DataFrame(input_data, index=[0])
+
             inputs = input_data.to_dict("records")
 
             funcs = [self.function] * len(inputs)
             kwargs = [self.function_kwargs] * len(inputs)
 
             output_data = self.executor.map(
-                safe_function1_for_map,
+                safe_function_for_map,
                 funcs,
                 inputs,
                 kwargs,
             )
 
-        return pd.DataFrame(output_data, index=input_data.index)
+        return pd.concat(
+            [input_data, DataFrame(output_data, index=input_data.index)], axis=1
+        )
 
-    def safe_function(self, *args, **kwargs):
+    def safe_function(self, *args, **kwargs) -> Dict:
         """
         Safely call the function, handling exceptions.
 
-        Note that this should not be submitted to fuu
+        Parameters
+        ----------
+        *args : tuple
+            Positional arguments to pass to the function.
+        **kwargs : dict
+            Keyword arguments to pass to the function.
+
+        Returns
+        -------
+        dict
+            The safe function outputs.
         """
         return safe_function(self.function, *args, **kwargs)
 
-    def submit(self, input: Dict):
-        """submit a single input to the executor
+    def submit(self, input: Dict) -> Future:
+        """
+        Submit a single input to the executor.
 
         Parameters
         ----------
         input : dict
+            The input dictionary to submit.
 
         Returns
         -------
-        Future  : Future object
+        Future
+            The Future object representing the submitted task.
         """
         if not isinstance(input, dict):
             raise ValueError("input must be a dictionary")
         # return self.executor.submit(self.function, input, **self.function_kwargs)
-        # Must call a function outside of the classs
+        # Must call a function outside of the class
         # See: https://stackoverflow.com/questions/44144584/typeerror-cant-pickle-thread-lock-objects
         return self.executor.submit(
             safe_function, self.function, input, **self.function_kwargs
         )
 
-    def submit_data(self, input_data: pd.DataFrame):
-        """submit dataframe of inputs to executor"""
+    def submit_data(self, input_data: pd.DataFrame) -> List[Future]:
+        """
+        Submit a dataframe of inputs to the executor.
+
+        Parameters
+        ----------
+        input_data : pd.DataFrame
+            The input data to submit.
+
+        Returns
+        -------
+        List[Future]
+            A list of Future objects representing the submitted tasks.
+        """
         input_data = pd.DataFrame(input_data)  # cast to dataframe for consistency
 
         if self.vectorized:
@@ -163,21 +227,62 @@ class Evaluator(BaseModel):
         return futures
 
 
-def safe_function1_for_map(function, inputs, kwargs):
+def safe_function_for_map(function: Callable, inputs: Dict, kwargs: Dict) -> Dict:
     """
     Safely call the function, handling exceptions.
+
+    Parameters
+    ----------
+    function : Callable
+        The function to call.
+    inputs : dict
+        The input dictionary to pass to the function.
+    kwargs : dict
+        The keyword arguments to pass to the function.
+
+    Returns
+    -------
+    dict
+        The safe function outputs.
     """
     return safe_function(function, inputs, **kwargs)
 
 
-def safe_function(function, *args, **kwargs):
+def safe_function(function: Callable, *args, **kwargs) -> Dict:
+    """
+    Safely call the function, handling exceptions.
+
+    Parameters
+    ----------
+    function : Callable
+        The function to call.
+    *args : tuple
+        Positional arguments to pass to the function.
+    **kwargs : dict
+        Keyword arguments to pass to the function.
+
+    Returns
+    -------
+    dict
+        The safe function outputs.
+    """
     safe_outputs = safe_call(function, *args, **kwargs)
     return process_safe_outputs(safe_outputs)
 
 
-def process_safe_outputs(outputs: Dict):
+def process_safe_outputs(outputs: Dict) -> Dict:
     """
-    Process the outputs of of safe_call, flattening the output.
+    Process the outputs of safe_call, flattening the output.
+
+    Parameters
+    ----------
+    outputs : dict
+        The outputs of safe_call.
+
+    Returns
+    -------
+    dict
+        The processed outputs.
     """
     o = {}
     error = False
@@ -202,11 +307,20 @@ def process_safe_outputs(outputs: Dict):
     return o
 
 
-def validate_outputs(outputs):
+def validate_outputs(outputs: DataFrame):
     """
     Looks for Xopt errors in the outputs and raises XoptError if found.
-    """
 
+    Parameters
+    ----------
+    outputs : DataFrame
+        The outputs to validate.
+
+    Raises
+    ------
+    XoptError
+        If any Xopt errors are found in the outputs.
+    """
     # Handles dicts or dataframes
     if not np.any(outputs["xopt_error"]):
         return
@@ -214,13 +328,16 @@ def validate_outputs(outputs):
     if "xopt_non_dict_result" in outputs:
         result = outputs["xopt_non_dict_result"]
         raise XoptError(
-            "Xopt evaluator returned a non-dict result, type is: "
-            f"{type(result)}, result is: {result}"
+            f"Xopt evaluator returned a non-dict result, type is: {type(result)}, result is: {result}"
         )
     else:
-        raise XoptError(
-            "Xopt evaluator caught an exception: " f"{outputs['xopt_error_str']}"
-        )
+        error_string = "Xopt evaluator caught exception(s):\n\n"
+        for i in range(len(outputs["xopt_error_str"])):
+            error_string += f"Evaluation index {i}:\n"
+            error_string += str(outputs["xopt_error_str"].iloc[i])
+            error_string += "\n"
+
+        raise XoptError(error_string)
 
 
 class DummyExecutor(Executor):
@@ -235,10 +352,51 @@ class DummyExecutor(Executor):
         self._shutdown = False
         self._shutdownLock = Lock()
 
-    def map(self, fn, *iterables, timeout=None, chunksize=1):
+    def map(self, fn: Callable, *iterables, timeout: float = None, chunksize: int = 1):
+        """
+        Map the function to the iterables.
+
+        Parameters
+        ----------
+        fn : Callable
+            The function to map.
+        *iterables : tuple
+            The iterables to map the function to.
+        timeout : float, optional
+            The timeout for the map operation. Defaults to None.
+        chunksize : int, optional
+            The chunk size for the map operation. Defaults to 1.
+
+        Returns
+        -------
+        map
+            The map object.
+        """
         return map(fn, *iterables)
 
-    def submit(self, fn, *args, **kwargs):
+    def submit(self, fn: Callable, *args, **kwargs) -> Future:
+        """
+        Submit a function to the executor.
+
+        Parameters
+        ----------
+        fn : Callable
+            The function to submit.
+        *args : tuple
+            The positional arguments to pass to the function.
+        **kwargs : dict
+            The keyword arguments to pass to the function.
+
+        Returns
+        -------
+        Future
+            The Future object representing the submitted task.
+
+        Raises
+        ------
+        RuntimeError
+            If the executor has been shut down.
+        """
         with self._shutdownLock:
             if self._shutdown:
                 raise RuntimeError("cannot schedule new futures after shutdown")
@@ -253,6 +411,14 @@ class DummyExecutor(Executor):
 
             return f
 
-    def shutdown(self, wait=True):
+    def shutdown(self, wait: bool = True):
+        """
+        Shut down the executor.
+
+        Parameters
+        ----------
+        wait : bool, optional
+            Whether to wait for the executor to shut down. Defaults to True.
+        """
         with self._shutdownLock:
             self._shutdown = True
