@@ -1,15 +1,22 @@
 from xopt.generators.bayesian.bayesian_exploration import BayesianExplorationGenerator
-from xopt.generators.bayesian.bayesian_generator import BayesianGenerator
-from xopt.generators.bayesian.expected_improvement import ExpectedImprovementGenerator
+from xopt.generators.bayesian.expected_improvement import (
+    ExpectedImprovementGenerator,
+    TDExpectedImprovementGenerator,
+)
 from xopt.generators.bayesian.mobo import MOBOGenerator
+from xopt.generators.bayesian.multi_fidelity import MultiFidelityGenerator
 from xopt.generators.bayesian.upper_confidence_bound import (
+    TDUpperConfidenceBoundGenerator,
     UpperConfidenceBoundGenerator,
 )
 
+
 __all__ = [
-    "BayesianGenerator",
     "BayesianExplorationGenerator",
     "MOBOGenerator",
     "UpperConfidenceBoundGenerator",
     "ExpectedImprovementGenerator",
+    "MultiFidelityGenerator",
+    "TDUpperConfidenceBoundGenerator",
+    "TDExpectedImprovementGenerator",
 ]
